//Eventhough the format of this file is .cpp no C++ functions have been used. 
//The .cpp format was only in order for Visual Studio to cope better with the file.

#include <stdafx.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <cv.h> //OpenCV
#include <highgui.h> //OpenCV
#include <cmath>
#include <string.h>
#include "dirent.h" //Directory /  File reader.

//These definitions cannot be changed in the settings-file.
#define DETECTIONLOGSIZE 200 //Max amount of faces to be logged, snapshotted and processed by AAM
#define MAXDETECTIONS 5 //Maximum amount of parallel detections
#define FRAMESKIP	10 //Amount of frames allowed between two faces, to call them the same.

//Detection struct
typedef struct {
	int x;
	int y;
	int width;
	int height;
	int score;
	int lastseen;
	int firstseen;
	int faceID;
	double aamscore;
	double aamminscore;
	int aamcount;
} detection_;

//Functions
void initDetections();
void shiftDetections();
void detect_and_draw(IplImage* image,double scale);
void saveSnapshot(detection_ face, IplImage* img);
IplImage* detect(IplImage* img);
void registerFaces(IplImage* img);
IplImage* drawFaces(IplImage* img);
double getFaceDist(detection_ X, detection_ Y);
double getFaceSizeChange(detection_ X, detection_ Y);
int findMoveNeighbor(detection_ face,int pos,IplImage* img);
double extractScore(const char* line);
int extractID(const char* line);
void evaluateAAM(FILE *detectionfile);
void loadSettings();
void readSettingLine(char* line,char* dest);
void readSettingInt(char* line,int* dest);
void readSettingDouble(char* line,double* dest);

//Variables
static CvMemStorage* storage = 0;
static CvHaarClassifierCascade* cascade = 0;
char *settingsfile = "Settings.rsf";

char AAM[256];
char CASCADE[256];


int MINSCORE; // = 15; //minimum amount of times the frame must have been spotted.
int MAXDIST; // = 30; //maximum distance a face is allowed to move pr. frame.
int MAXSIZEJUMP; // = 20; //Max amount of change in face size
int MINFACESIZE; // = 120;  //Minimum size of a face (pixels)
double SEARCHSCALING; // = 1.2; //Incrementation of searchwindow-scale. (1.1 = 10% increase)
int SNAPSHOTINTERVAL; // = 10; //Amount of scorepoints between every snapshot
int MAXSNAPSHOTS; // = 5; //Maximum amount of snapshots taken per face.
int OUTPUTSIZE; // = 200; //Size of the faceimages...
double AAMTHRESHOLD; // = 1.2; //Amount a sample is allowed to deviate from the Active Appearance Model.
int AAMMINSEARCH; //Lowest score search-mode
int SHOWVIDEO; //Display video while processing
int COLORMODE;	//Color mode or greyscale

int currentframe = 0;
int facedetections = 0;

detection_ detections[MAXDETECTIONS][FRAMESKIP]; //Array of detections vs time.
detection_ faces[DETECTIONLOGSIZE]; //Collection of accepted faces


//MAIN PROGRAM..........................................................................
int main( int argc, char** argv ) {
	bool keepopen = false;
	char str[50] = {0};
	char* input = argv[1];
    struct dirent *d;
    DIR *dir;
    char buf[256];
	char systemcallstring[256] = "";
	storage = cvCreateMemStorage(0);	
	FILE *detectionfile;
	

	if(!input){
		printf("-------------- Face Detection and Recognition -----------\n");
		printf("---------------------------------------------------------\n");
		printf("Welcome!\n");
		printf("Please enter input file location:\n");
		gets(str);
		printf("\n# Loading data from: %s.\n",str);
		input = str;
		keepopen = true;
	}
	printf("  Done.\n\n");

	printf("# Loading Settings...\n");
	loadSettings();
	printf("  Done.\n\n");

	cascade = (CvHaarClassifierCascade*)cvLoad( CASCADE,0, 0, 0 );
	initDetections();
	
	printf("# Emptying detection directory.\n");
	dir = opendir("processing");
	if(!dir){
		printf("  ERROR!\n  The folder /processing does not exist! \n  Please create this folder and restart the software!\n\n\n");	
	}else{
		while(d = readdir(dir)){
				sprintf(buf, "%s/%s", "processing", d->d_name);
				remove(buf);
		}
		remove("Detections.txt");
		printf("  Done.\n\n");
	}
	

	printf("# Starting ViolaJones face-detection. \n");
	printf("  Using cascade: %s \n",CASCADE);
	
	CvCapture* capture = cvCreateFileCapture(input);
	IplImage* frame;
	while(1) {
		frame = cvQueryFrame( capture );
		if( !frame ) break;
		currentframe++;
		registerFaces(frame);
		if(SHOWVIDEO){
			frame = drawFaces(frame);
			cvShowImage( "Displaying Video...", frame );
			char c = cvWaitKey(1);
			if( c == 27 ) break;
		}
	}
	cvReleaseCapture( &capture );
	cvDestroyWindow( "Displaying Video..." );

	

	printf("  Done.\n\n");
	printf("# Total unique face detections: %d\n\n",facedetections);

	printf("# Starting AAM recognition.\n");
	printf("  Using Appearance Model: %s.\n",AAM);
	printf("  Mode set to: ",AAM);
	
	if(COLORMODE){
		printf("Color.\n");
		strcat(systemcallstring,"aamcm.exe s "); //Color AAM software	
	}else{
		printf("Grayscale.\n");
		strcat(systemcallstring,"aamc.exe s "); //Grayscale AAM software
	}
	strcat(systemcallstring,AAM);
	strcat(systemcallstring," processing > Detections.txt");

	//"aamcm.exe s model.amf processing > Detections.txt"
	system(systemcallstring);
	printf("  Done.\n\n");

	
	printf("----------------- Processing completed  -----------------\n");
	printf("---------------------------------------------------------\n");

	detectionfile = fopen("Detections.txt", "r"); //Load the aam datalog
	evaluateAAM(detectionfile); //Extract numbers from log and assign to face-array.

	printf("---------------------------------------------------------\n");
	printf("---------------------------------------------------------\n");


	if(keepopen){
		printf("\nPress enter to exit...");
		gets(str);
	}
	return (0);
}
//......................................................................................



//Functions ----------------------------------------------------------------------------
//--------------------------------------------------------------------------------------

//Detect faces in the video
void registerFaces(IplImage* img){

	// IMAGE PREPARATION:
	//
	IplImage* gray = cvCreateImage( cvSize(img->width,img->height), 8, 1 );
	cvCvtColor( img, gray, CV_BGR2GRAY );
	//cvEqualizeHist( small_img, small_img );
	
	// DETECT OBJECTS IF ANY
	//
	cvClearMemStorage(storage);
	CvSeq* objects = cvHaarDetectObjects(
		gray,
		cascade,
		storage,
		SEARCHSCALING,
		2,
		0 /*CV_HAAR_DO_CANNY_PRUNING*/,
		cvSize(MINFACESIZE, MINFACESIZE)
	);

	shiftDetections(); //Shifting the detection content 1 frame.. (incrementing time)
	for(int i = 0; i < (objects ? objects->total : 0); i++ ) {
		
		CvRect* r = (CvRect*)cvGetSeqElem( objects, i );
		
		//Initiate a face struckt.
		detection_ thisface;
		thisface.x = r->x;  //Returned x-position of face from the haardetection
		thisface.y = r->y;	//Returned y-position of...
		thisface.width = r->width; //Returned width of...
		thisface.height = r->height; //Returned height of...
		thisface.firstseen = currentframe;
		thisface.lastseen = currentframe;
		thisface.score = 0;
		thisface.faceID = -1;
		thisface.aamscore = 0;
		thisface.aamminscore = 999;
		thisface.aamcount = 0;

		if(i<MAXDETECTIONS){
			if(findMoveNeighbor(thisface,i,img)){
				//printf("Found a dublicate, score incremented..\n");
			}else{
				//printf("New Face added.\n");
			}
		}
	}
	cvReleaseImage( &gray );
	
}

//Calculate distance between two faces
double getFaceDist(detection_ X, detection_ Y){
	double xchange = (double)(X.x-Y.x);
	double ychange = (double)(X.y-Y.y);
	double dist = sqrt(xchange*xchange+ychange*ychange);
	//double dist = xchange+ychange;
	return dist;
}

//Calculate size distance between two faces
double getFaceSizeChange(detection_ X, detection_ Y){
	double hchange = (double)(X.height-Y.height);
	double wchange = (double)(X.width-Y.width);
	double dist = sqrt(hchange*hchange+wchange*wchange);
	//double dist = (hchange+wchange)*0.5;
	return dist;
}

//Do the time-continous search between faces.
int findMoveNeighbor(detection_ face,int pos,IplImage* img){
	int found = 0;
	int oldscore;
	int oldFaceID;
	int oldFirstSeen;
//printf("searching: ");
	for(int i = 0; i < MAXDETECTIONS && !found; i++ ) {
		for(int j = 0; j < FRAMESKIP-1  && !found; j++){
			//Check if the cell contains data and if so, if the face is within proximity
			if(detections[i][j].x != -1){
				if(getFaceDist(face,detections[i][j]) < MAXDIST){
//printf("Found pos match...\n");
					if(getFaceSizeChange(face,detections[i][j])< MAXSIZEJUMP){
						detections[pos][FRAMESKIP-1] = detections[i][j]; //Bring to the front
						oldscore = detections[pos][FRAMESKIP-1].score; //save the old score
						oldFaceID = detections[pos][FRAMESKIP-1].faceID; //save the old ID
						oldFirstSeen = detections[pos][FRAMESKIP-1].firstseen; //save the old FirstSeen

						detections[pos][FRAMESKIP-1] = face; //Update coords etc to the new.
						detections[pos][FRAMESKIP-1].score = oldscore+1; //Keep the old score and increment
						detections[pos][FRAMESKIP-1].faceID = oldFaceID; //Keep the old faceID
						detections[pos][FRAMESKIP-1].firstseen = oldFirstSeen; //Keep the old firstseen framestamp.

						detections[i][j].x = -1; //Mark old cell as empty
						
						if(detections[pos][FRAMESKIP-1].score >= MINSCORE){ //Has passed the threshold and will be saved
							if(detections[pos][FRAMESKIP-1].faceID == -1){//Is it a new addition?
								facedetections++;
								detections[pos][FRAMESKIP-1].faceID = facedetections-1; //Cell nbr in the faces array
								faces[facedetections-1] = detections[pos][FRAMESKIP-1]; //Save in faces array
								
								//[Save a snapshot in processing/]
								saveSnapshot(faces[facedetections-1],img);

								//printf("New face added to facelog\n");
							}else if(detections[pos][FRAMESKIP-1].score > MINSCORE){  //A face needs update
								faces[detections[pos][FRAMESKIP-1].faceID] = detections[pos][FRAMESKIP-1]; //Update in faces array
								
								if((detections[pos][FRAMESKIP-1].score-MINSCORE) % SNAPSHOTINTERVAL == 0 && (detections[pos][FRAMESKIP-1].score-MINSCORE)/SNAPSHOTINTERVAL < MAXSNAPSHOTS){ //is it dividable with snapshotinterval (ie. save on 10,15,20,25 when MINSCORE is 10 and has there been taken more than max allowed?)
									saveSnapshot(faces[detections[pos][FRAMESKIP-1].faceID],img); //Then save another snapshot								
								}
								//printf("Updated facelog: %d \n",currentframe);
							}

						}			
						
						found = 1; //Exit the loop

					}
				}
			}
		}
	}
//printf("\n");
	if(!found){
		detections[pos][FRAMESKIP-1] = face; //If no simillar faces found, just create a new.
	}
	return found;
}

//Shift the time-detection array 1 step, ie. increment time by 1 frame.
void shiftDetections(){
	for(int i = 0; i < MAXDETECTIONS; i++ ) {
		for(int j = 0; j < FRAMESKIP-1; j++){
			detections[i][j] = detections[i][j+1]; //Shift value one left.
		}
		detections[i][FRAMESKIP-1].x = -1; //Empty the value of the most right.
	}
}

//Initialize detection array.
void initDetections(){
	for(int i = 0; i < MAXDETECTIONS; i++ ) {
		for(int j = 0; j < FRAMESKIP; j++){
			detections[i][j].x = -1;
		}
	}
}

//Draw all accepted faces in the frame
IplImage* drawFaces(IplImage* img){
for(int i = 0; i < facedetections; i++ ) {
	if(faces[i].lastseen == currentframe){
		cvRectangle(
				img,
				cvPoint(faces[i].x,faces[i].y),
				cvPoint(faces[i].x+faces[i].width,faces[i].y+faces[i].height),
				CV_RGB(255,0,0)
		);
	}
}
return img;
}

//Save a snapshow of the face
void saveSnapshot(detection_ face, IplImage* img){
	int id = face.faceID;
	int nbr = (face.score-MINSCORE)/SNAPSHOTINTERVAL;
	char sid[5];
	char snbr[5];
	char filename[25] = "processing/tmp";

	itoa(id,sid,10);
	itoa(nbr,snbr,10);

	strcat(filename,sid);
	strcat(filename,"_");
	strcat(filename,snbr);
	strcat(filename,".bmp");
	
	//Scaling and translation should be implemented as settings-variables.
	cvSetImageROI(img, cvRect(face.x-0.1*face.width,(int)face.y+0.05*face.height,face.width*1.2,face.height*1.2)); //Mark region to be saved.. The area is scaled and translated a bit to help AAM..
	IplImage *cropped = cvCreateImage(cvSize(OUTPUTSIZE,OUTPUTSIZE),img->depth,img->nChannels);
	cvResize( img, cropped, CV_INTER_LINEAR );
	cvSaveImage(filename,cropped);
	cvResetImageROI(img);
	cvReleaseImage( &cropped );
}

//Extract score from the AAM text-file.
double extractScore(const char* line){
	char part[10];
	part[0] = line[33];
	part[1] = line[34];
	part[2] = line[35];
	part[3] = line[36];
	part[4] = line[37];
	part[5] = line[38];
	part[6] = line[39];
	return atof(part);
}

//Extract ID from the AAM text-file
int extractID(const char* line){
	char part[10];
	int span;
	int dist = 0;


	int len = strlen(line);
	for(int i = 0; i < len; i++)
	{
		if(line[i] == '_'){
			dist = i;
			break;
		}
	}
	span = dist-52;  //52 chars to the start of the ID..
	for(int i = 0; i<span;i++){
		part[i] = line[52+i];
	}

	return atof(part);
}

//Run through the AAM text-file.
void evaluateAAM(FILE *detectionfile){
	char aam[256];
	int aamline = 0;
	int subjectid[DETECTIONLOGSIZE];
	int subjectcount = 0;
	int id;
	int current;
	double score;

	//Run through detectionfile and add that scores to their respective faces.
	while(fgets(aam,256,detectionfile)){
		aamline++;
		if(!((aamline-1)%3)){ //Every 3rd line from line nr.1 contains Face ID..
			id = extractID(aam);
		}
		else if(!((aamline-2)%3)){ //Every 3rd line from line nr.2 contains score..
			score = extractScore(aam);
			if(score > 0){
				faces[id].aamcount++;
				faces[id].aamscore += score;
				if(score < faces[id].aamminscore){
					faces[id].aamminscore = score;
				}
			}
		}
	}
	
	//Run through all faces and evaluate their total scores (averaged).
	for(int i = 0;i<facedetections;i++){
		
		if(faces[i].aamcount>0){
			
			if(AAMMINSEARCH){ //Minimum mode
				if(faces[i].aamminscore < AAMTHRESHOLD){
					subjectid[subjectcount] = i;
					subjectcount++;
				}
			}else{ //Average mode
				if(faces[i].aamscore/faces[i].aamcount < AAMTHRESHOLD){
					subjectid[subjectcount] = i;
					subjectcount++;

				}
			}
		}
	}

	//Run through array of faces that have passed the AAM threshold i.e the actual detections of the subject..
	if(subjectcount > 0){
		printf("The subject has been recognized. \nSubject is approximately present in:\n\n");
		for(int i = 0;i<subjectcount;i++){
			current = subjectid[i];
			printf("-Frames: %d - %d",faces[current].firstseen,faces[current].lastseen);
			printf(" (%2.2f s - %2.2f s)\n",(faces[current].firstseen)*0.04,(faces[current].lastseen)*0.04);
			printf(" Based on %d AAM evaluation(s)  -  (ID: %d, Mah.Dist.: ",faces[current].aamcount,current);
			if(AAMMINSEARCH){ //Minimum mode
				printf("%2.2f)\n",faces[current].aamminscore);
			}else{ //Average mode
				printf("%2.2f)\n",faces[current].aamscore/faces[current].aamcount);
			}
		}
		printf("\n\nTime estimates are based on a framerate at 25 pr. sec.\n");
	}else{
		printf("The subject could not be recognized.\n\n");
	}
}

//Load the settings.
void loadSettings(){
	char setting[256];
	int settingline = 0;

	FILE *file = fopen(settingsfile, "r");

	//Run through detectionfile and add that scores to their respective faces.
	while(fgets(setting,256,file)){
		settingline++;
		
		switch(settingline){
			case 5:
				readSettingLine(setting,CASCADE);	
			break;

			case 7:
				readSettingLine(setting,AAM);		
			break;

			case 9:
				readSettingDouble(setting,&SEARCHSCALING);		
			break;

			case 11:
				readSettingInt(setting,&MINSCORE);		
			break;

			case 13:
				readSettingInt(setting,&MAXDIST);		
			break;

			case 15:
				readSettingInt(setting,&MAXSIZEJUMP);		
			break;

			case 17:
				readSettingInt(setting,&MINFACESIZE);		
			break;

			case 19:
				readSettingInt(setting,&SNAPSHOTINTERVAL);		
			break;

			case 21:
				readSettingInt(setting,&MAXSNAPSHOTS);		
			break;

			case 23:
				readSettingInt(setting,&OUTPUTSIZE);		
			break;

			case 25:
				readSettingDouble(setting,&AAMTHRESHOLD);		
			break;

			case 27:
				readSettingInt(setting,&AAMMINSEARCH);		
			break;

			case 29:
				readSettingInt(setting,&SHOWVIDEO);		
			break;

			case 31:
				readSettingInt(setting,&COLORMODE);		
			break;
		}
	}
}

//Read a text-line from the settings-file.
void readSettingLine(char* line, char* dest){
	int len = strlen(line);
	int dist;
	char part[255]; 
	
	for(int i = 0; i < len; i++)
	{
		if(line[i] == ';'){
			dist = i;
			break;
		}
	}
	for(int i = 0; i<dist;i++){
		part[i] = line[i];
	}
	part[dist] = '\0';

	strncpy(dest, part, 255);
}

//Read an int from the settings-file.
void readSettingInt(char* line, int* dest){
	int len = strlen(line);
	int dist;
	char part[255]; 
	
	for(int i = 0; i < len; i++)
	{
		if(line[i] == ';'){
			dist = i;
			break;
		}
	}	
	for(int i = 0; i<dist;i++){
		part[i] = line[i];
	}
	part[dist] = '\0';

	*dest = atoi(part);
	//printf("String: %s ; Computed: %d\n",part,atoi(part));
}

//Read a double from the settings-file.
void readSettingDouble(char* line, double* dest){
	int len = strlen(line);
	int dist;
	char part[255]; 
	
	for(int i = 0; i < len; i++)
	{
		if(line[i] == ';'){
			dist = i;
			break;
		}
	}
	for(int i = 0; i<dist;i++){
		part[i] = line[i];
	}
	part[dist] = '\0';

	*dest = atof(part);
	//printf("String: %s ; Computed: %f\n",part,atof(part));
}